#!/usr/bin/env python
# -*- coding: utf-8 -*- 

from pygments.lexer import this, Lexer, RegexLexer
from pygments.token import *
from pygments.style import Style
from pygments.filter import simplefilter
import re

class HahaLexer(RegexLexer):
    name = 'HAHA'
    aliases = ['haha']
    filenames = ['*.haha']

    keywords = ['ARRAY', 'BOOLEAN', 'INT', 'Z', 'axiom', 'begin', 'counter', 'do', 'else', 'end', 'function', 'if', 'invariant', 'postcondition', 'precondition', 'predicate', 'skip', 'then', 'var', 'while']
    text_operators = ['and', 'exists', 'forall', 'mod', 'not', 'or', 'pattern']
    operators = ['!=', '*', '+', '-', '->', '/\\', '/', '¬', '<', '<->', '<=', '<>', '=', '>', '>=', '\\/', '^', '→', '↔', '⇒', '⇔', '∀', '∃', '∧', '∨', '≠', '≤', '≥']
    punctuation = ['(', ')', ',', ':', ';', '[', ']']
    literals = ['true', '⊤', 'false', '⊥']

    ident = r'[_a-zA-Z][_a-zA-Z0-9]*'
    number = r'([1-9][0-9]*)|0'
    #escape = "\\\\[btnfr\"'\\]"
    escape = r'\\.'

    tokens = {
      'root': [
          (r'//(.*?)\n', Comment.Single),
          (r'/(\\\n)?[*](.|\n)*?[*](\\\n)?/', Comment.Multiline),
          ('|'.join([re.escape(kwd) + '\\b' for kwd in keywords]), Keyword),
          ('|'.join([re.escape(op) + '\\b' for op in text_operators]), Operator.Word),
          ('|'.join([re.escape(op) for op in operators]), Operator),
          ('|'.join([re.escape(symbol) for symbol in punctuation]), Punctuation),
          ('|'.join([re.escape(lit) + '\\b' for lit in literals]), Literal),
          (r'[{}]', Keyword), # hack...
          (ident, Name),
          (number, Number),
          (r'\s+', Text),
          (r'"', String, 'string-single'),
          (r'"', String, 'string-double'),
       ],
       'string-double': [
          (r'"', String, '#pop'),
          (escape, String.Escape),
          (r'[^\"]+', String)
       ],
       'string-single': [
          ("'", String, '#pop'),
          (escape, String.Escape),
          ("[^\\']+", String)
       ],
    }

@simplefilter
def assertion_filter(self, lexer, stream, options):
    inside_assertion = False
    for ttype, value in stream:
        if value == '{': inside_assertion = True
        if inside_assertion: 
            yield ttype.Alt, value
        else: 
            yield ttype, value
        if value == '}': inside_assertion = False

def fix_style(style):
    new_styles = {}
    for token in style.styles:
        new_styles[token] = style.styles[token]
        if str(token).endswith('.Alt'): continue
        alt_token = token.Alt
        if style.style_for_token(token)['italic']:
            new_styles[alt_token] = 'noitalic'
        else:
            new_styles[alt_token] = 'italic'
    class FixedStyle(Style):
        title="?"
        background_color = style.background_color
        default_style = style.default_style
        styles = new_styles
    return FixedStyle

