package net.afyre.haha.demo.solver.smt;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.io.OutputStreamWriter;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.FutureTask;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import net.afyre.haha.demo.solver.smt.sexpr.DefaultSExprFactory;
import net.afyre.haha.demo.solver.smt.sexpr.SExpr;
import net.afyre.haha.demo.solver.smt.sexpr.SExprParser;
import net.afyre.haha.demo.solver.smt.sexpr.SExprPrinter;

/**
 * Handle used to manage the solver process (and communicate with it).
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class SolverProcess {
  /** External process parameters. */
  private final ProcessBuilder processBuilder;
  
  /** External process handle. */
  private Process process;
  
  /** Logger used to display error messages. */
  private final SmtLogger logger;
  
  /** Flag set when the external process is started. */
  private boolean running;
  
  /** Pipe connected to stdin of the external process. */
  private BufferedWriter stdin;

  /** Pipe connected to stdout of the external process. */
  private BufferedReader stdout;

  /** Pipe connected to stderr of the external process. */
  private BufferedReader stderr;
  
  /** Thread used to service stderr of the external process. */
  private StderrHandler stderrHandler;
  
  /** Object used to convert S-Expressions to text. */
  private final SExprPrinter exprPrinter = new SExprPrinter();
  
  /** Object used to parse expressions returned by the solver. */
  private final SExprParser exprParser = 
      new SExprParser(new DefaultSExprFactory());
  
  /**
   * Constructor.
   * @param logger Logger used to display error messages.
   * @param executable Path to the solver executable.
   * @param arguments Command line arguments.
   */
  public SolverProcess(final SmtLogger logger,
                       final String executable, 
                       final String... arguments) {   
    this.logger = logger;
    
    // Build the command
    final String[] cmd = new String[arguments.length + 1];
    cmd[0] = executable;
    for (int i = 0; i < arguments.length; i++) {
      cmd[i + 1] = arguments[i];
    }
    processBuilder = new ProcessBuilder(cmd);
  }
  
  /**
   * Start the solver if it is not already running.
   * @return true on success, false otherwise.
   */
  public boolean start() {
    if (!running) {
      try {
        process = processBuilder.start();
      } catch (final IOException e) {
        logger.error("Unable to start the solver", e);
      }
      stdin = new BufferedWriter(new OutputStreamWriter(
          process.getOutputStream()));
      stdout = new BufferedReader(new InputStreamReader(
          process.getInputStream()));
      stderr = new BufferedReader(new InputStreamReader(
          process.getErrorStream()));
      stderrHandler = new StderrHandler();
      stderrHandler.start();
      running = true;
    }
    return true;
  }
  
  /**
   * Send an expression to the solver.
   * 
   * @param cmd Expression to be sent.
   * @return true on success, false otherwise.
   */
  public boolean send(final SExpr cmd) {
    boolean result;
    final String text = cmd.doSwitch(exprPrinter);
   
    if (!running) { 
      logger.error("Solver is not running.");
      result = false; 
    } else {
      try {
        stdin.write(text);
        stdin.write("\n");
        stdin.flush();
        logger.input(text);
        result = true;
      } catch (final InterruptedIOException e) {
        logger.error("I/O interrupted)");
        Thread.currentThread().interrupt();
        result = false;
      } catch (final IOException e) {
        logger.error("Cannot communicate with solver.", e);
        result = false;
      }
    }
    return result;
  }
    
  /**
   * Read single expression printed by the solver.
   * 
   * Note that solver should be restarted whenever 
   * this method returns null, as it might be in
   * an incosistent state.
   * 
   * @param hardTimeout Timeout for the read operation.
   * @param unit Unit in which the timeout is expressed.
   * 
   * @return Expression or null (in case of an error or read timeout).
   */
  public SExpr read(final int hardTimeout, final TimeUnit unit) {
    SExpr result;
    
    start();

    final Callable<SExpr> readProc = 
        new Callable<SExpr>() {
          @Override
          public SExpr call() throws Exception {
            return exprParser.parse(stdout);
          }
        };
    final FutureTask<SExpr> readTask = 
        new FutureTask<SExpr>(readProc);
    Executors.newSingleThreadExecutor().execute(readTask);

    try {
      result = readTask.get(hardTimeout, unit);
    } catch (final InterruptedException e) {
      result = null;
      logger.warning("Wait interrupted.");
      readTask.cancel(true);
      Thread.currentThread().interrupt();
    } catch (final ExecutionException e) {
      logger.error("Read error", e);
      result = null;
    } catch (final TimeoutException e) {
      logger.error("Solver timeout");
      result = null;
    }
    
    if (result != null) {
      logger.output(result.toString());
    }
    return result;
  }

  /**
   * Stop the solver.
   * @param timeout Timeout after which the external process is killed.
   * @param unit Unit in which the timeout is expressed.
   */
  public void stop(final int timeout, final TimeUnit unit) {
    int ret;

    final Callable<Integer> waitProc = 
        new Callable<Integer>() {
          @Override
          public Integer call() throws Exception {
            return process.waitFor();
          }
        };
    final FutureTask<Integer> waitTask = new FutureTask<Integer>(waitProc);
    Executors.newSingleThreadExecutor().execute(waitTask);
    
    try {
      ret = waitTask.get(timeout, unit);
      // CHECKSTYLE:OFF
    } catch (final Exception e) {
      // CHECKSTYLE:ON
      logger.warning("Unable to stop solver");
      process.destroy();
      running = false;
      return;
    }
    if (ret == 0) {
      logger.info(String.format("Solver exited with code %d", ret));
    } else {
      logger.error(String.format("Solver exited with code %d", ret));
    }
    running = false;
  }
  
  /**
   * Is the external process started?
   * @return .
   */
  public boolean isRunning() {
    return running;
  }
  
  /**
   * Thread that reads lines from stderr.
   */
  private class StderrHandler extends Thread {
    @Override
    public void run() {
      boolean finished = false;
      
      while (!finished) {
        final String line;
        
        try {
          line = stderr.readLine();
        } catch (final InterruptedIOException e) {
          return;
        }  catch (final IOException e) {
          return;
        }
        if (line == null) {
          finished = true;
        } else {
          if (!"".equals(line.trim())) {
            logger.warning(String.format("Solver error: %s", line)); 
          }
        }
      }
    }
  }
}
