package net.afyre.haha.demo.solver.smt.sexpr;

import java.math.BigDecimal;
import java.math.BigInteger;

/**
 * Interface of objects used to inspect S-Expressions.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 * @param <T> Type of values returned by the switch methods.
 */
public interface SExprSwitch<T> {
  /**
   * Inspect an integer numeral.
   * @param value Integer value (non-negative).
   * @return Inspection result.
   */
  T caseInteger(BigInteger value);  
  
  /**
   * Inspect a decimal numeral.
   * @param value Decimal value (non-negative).
   * @return Inspection result.
   */
  T caseDecimal(BigDecimal value);
  
  /**
   * Inspect a hexadecimal numeral.
   * @param value Numeral value (non-negative).
   * @param length Number of digits (including leading zeros)
   * @return Inspection result.
   */
  T caseHex(BigInteger value, int length);
  

  /**
   * Inspect a binary numeral.
   * @param value Numeral value (non-negative).
   * @param length Minimal number of digits.
   *               Can be used to introduce leading zeros. 
   * @return Inspection result.
   */
  T caseBin(BigInteger value, int length);
  
  /**
   * Inspect a quoted string.
   * @param value String text.
   * @return Inspection result.
   */
  T caseString(String value);
  
  /**
   * Inspect a symbol.
   * @param name Symbol name.
   * @return Inspection result.
   */
  T caseSymbol(String name);

  /**
   * Inspect a keyword.
   * @param name Keyword name (without leading colon).
   * @return Inspection result.
   */
  T caseKeyword(String name);
  
  /**
   * Inspect a list.
   * @param elements List elements.
   * @return Inspection result.
   */
  T caseList(Iterable<SExpr> elements);
}
