package net.afyre.haha.demo.solver.context;

import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;

import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;
import net.afyre.haha.demo.solver.expr.util.ExprPrinter;

/**
 * Declaration visitor used to convert declarations to text.
 *
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class DeclarationPrinter implements DeclarationVisitor {
  /** Output stream. */
  private final Writer output;
  
  /** Object used to print expressions. */
  private final ExprPrinter exprPrinter;
  
  /**
   * Create a visitor which will write all 
   * inspected objects to an output stream.
   * @param output Output stream.
   */
  public DeclarationPrinter(final Writer output) {
    this.output = output;
    this.exprPrinter = new ExprPrinter(output);
  }
  
  // TODO: Move this to utils
  /**
   * Print an object (using toString()) to the output stream, 
   * converting any IOException to RuntimeExceptions.
   * @param obj Object to be printed (can be null).
   */
  private void print(final Object obj) {
    final String message;
    
    if (obj == null) {
      message = "null";
    } else {
      message = obj.toString();
    }
    try {
      output.write(message);
    } catch (final IOException e) {
      throw new RuntimeException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void visitAxiom(final String name,
                         final Formula axiom) {
    print("AXIOM ");
    print(name);
    print(" = ");
    axiom.accept(exprPrinter);
  }

  /** {@inheritDoc} */
  @Override
  public void visitTermDeclaration(final String name, 
                                final ValueType type) {
    print("VAR ");
    print(name);
    print(" : ");
    type.accept(exprPrinter);
  }
   
  /** {@inheritDoc} */
  @Override
  public void visitTermDefinition(final String name, 
                                  final ValueType type, 
                                  final Term value) {
    print("CONST ");
    print(name);
    print(" : ");
    type.accept(exprPrinter);
    print(" = ");
    value.accept(exprPrinter);
  }

  /** {@inheritDoc} */
  @Override
  public void visitPropDeclaration(final String name) {
    print("VAR ");
    print(name);
    print(" : ");
    print("Prop");
  }
  
  /** {@inheritDoc} */
  @Override
  public void visitPropDefinition(final String name, 
                                  final Formula value) {
    print("CONST ");
    print(name);
    print(" = ");
    value.accept(exprPrinter); 
  }

  /** {@inheritDoc} */
  @Override
  public void visitFunctionDeclaration(final String name,
                                       final Iterable<VariableBinding> args,
                                       final ValueType result) {
    print("UNINTERPRETED_FUNCTION " + name + "(");
    
    visitFunctionDefinition(name, args, result, null);
  }
  
  /** {@inheritDoc} */
  @Override
  public void visitFunctionDefinition(final String name,
                                      final Iterable<VariableBinding> args,
                                      final ValueType result, 
                                      final Term body) {
    print("FUNCTION " + name + "(");
    
    final Iterator<VariableBinding> iter = args.iterator();
    while (iter.hasNext()) {
      final VariableBinding arg = iter.next();
      print(arg.getName());
      print(" : ");
      arg.getType().accept(exprPrinter);
      if (iter.hasNext()) {
        print(", ");
      }
    }
    print(") : ");
    result.accept(exprPrinter);
    if (body == null) {
      print(";");
    } else {
      print(" = ");
      body.accept(exprPrinter);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void visitPredicateDeclaration(
      final String name,
      final Iterable<VariableBinding> args) {
    visitPredicateDefinition(name, args, null);
  }
  
  /** {@inheritDoc} */
  @Override
  public void visitPredicateDefinition(final String name,
                                       final Iterable<VariableBinding> args,
                                       final Formula body) {
    print("PREDICATE " + name + "(");
    
    final Iterator<VariableBinding> iter = args.iterator();
    while (iter.hasNext()) {
      final VariableBinding arg = iter.next();
      print(arg.getName());
      print(" : ");
      arg.getType().accept(exprPrinter);
      if (iter.hasNext()) {
        print(", ");
      }
    }
    print(")");
    if (body == null) {
      print(";");
    } else {
      print(" = ");
      body.accept(exprPrinter);
    }
  }
}
