package net.afyre.haha.demo.solver.expr.util;

import java.math.BigInteger;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.ExpressionSwitch;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * An abstract expression switch that returns a default value
 * for all cases.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 * @param <Result> Type of values computed by the switch.
 */
public abstract class DefaultExpressionSwitch<Result> 
    implements ExpressionSwitch<Result> {
  /**
   * Get the default value.
   * @return Default value.
   */
  protected abstract Result getDefault();

  /** {@inheritDoc} */
  @Override
  public Result caseIntConst(final BigInteger value) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVariable(final String var) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseMkTuple(final Term... elements) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseTupleAccess(final Term tuple, final int index) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseUminus(final Term arg) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseAddition(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseSubtraction(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseMultiplication(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseDivision(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseRemainder(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result casePower(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseLet(final String var, final ValueType type, 
                        final Term value, final Term body) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseLetProp(final String var, 
                            final Formula value, 
                            final Term body) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseIfThenElse(final Formula cond, 
                               final Term ifTrue, 
                               final Term ifFalse) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseArrayAccess(final Term arr, final Term index) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseArrayUpdate(final Term arr, 
                                final Term index, 
                                final Term value) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseCall(final String func, final Term... args) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseConcat(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseSelectBits(final Term vec, 
                               final int left, 
                               final int right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseShl(final Term vec, final int offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseShr(final Term vec, final int offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseSX(final Term vec, final int offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseZeroExtend(final Term vec, final int offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseRepeat(final Term vec, final int count) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseRotateLeft(final Term vec, final int offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseRotateRight(final Term vec, final int offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorAnd(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorOr(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorXor(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorNand(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorNor(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorXnor(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorCmp(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorSum(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorDiff(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorUminus(final Term vec) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorProd(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorShl(final Term vec, final Term offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorLshr(final Term vec, final Term offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorRshr(final Term vec, final Term offset) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorUdiv(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorSdiv(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorUrem(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorSrem(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorSmod(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorLt(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorGt(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorLe(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorGe(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorSlt(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorSgt(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorSle(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseVectorSge(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseTrue() {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseFalse() {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseFormulaVariable(final String name) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseNegation(final Formula arg) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseConjunction(final Formula left, final Formula right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseDisjunction(final Formula left, final Formula right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseImplication(final Formula left, final Formula right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseXor(final Formula left, final Formula right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseEquivalence(final Formula left, final Formula right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseEq(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseNeq(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseDistinct(final Term... args) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseLt(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseGt(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseLe(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseGe(final Term left, final Term right) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseExists(final Iterable<VariableBinding> variables, 
                           final Formula arg,
                           final Expression... patterns) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseForall(final Iterable<VariableBinding> variables, 
                           final Formula arg,
                           final Expression... patterns) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseFLet(final String var, 
                         final ValueType type, 
                         final Term value, 
                         final Formula body) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseFLetProp(final String var, 
                             final Formula value, 
                             final Formula body) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result casePredicateCall(final String pred, final Term... args) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseRational() {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseInt() {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseBitvector(final int size) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseArray(final ValueType indexType, 
                          final ValueType valueType) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseTuple(final ValueType... types) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseProp() {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result caseFunction(final ValueType resultType, 
                             final ValueType... argTypes) {
    return getDefault();
  }

  /** {@inheritDoc} */
  @Override
  public Result casePredicate(final ValueType... argTypes) {
    return getDefault();
  }
}
