package net.afyre.haha.demo.solver.expr.util;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.ExpressionFactory;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.formulae.FormulaSwitch;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.terms.TermSwitch;
import net.afyre.haha.demo.solver.expr.types.AtomicType;
import net.afyre.haha.demo.solver.expr.types.NonValueType;
import net.afyre.haha.demo.solver.expr.types.StructuredType;
import net.afyre.haha.demo.solver.expr.types.Type;
import net.afyre.haha.demo.solver.expr.types.TypeSwitch;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * A switch which applies itself recursively to reconstruct an expression.
 * It is useful as a superclass for switches which need to
 * change some parts of an expression.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class ExprTransformer 
    implements TermSwitch<Term>, 
               FormulaSwitch<Formula>, 
               TypeSwitch<Type> {
  /** Factory used to create expressions. */
  private final ExpressionFactory factory;
  
  /**
   * Constructor.
   * @param factory Factory which will be used to create new expressions.
   * @throws NullPointerException If the factory is null. 
   */
  public ExprTransformer(final ExpressionFactory factory) {
    if (factory == null) {
      throw new NullPointerException("factory");
    }
    this.factory = factory;
  }
  
  /**
   * Get the factory used to create expressions.
   * @return Factory object.
   */
  protected ExpressionFactory getFactory() {
    return factory;
  }
  
  /**
   * Apply this switch to an expression.
   * @param expr Target expression.
   * @return Transformation result.
   */
  private Expression applyToExpression(final Expression expr) {
    final Expression result;
    if (expr instanceof Type) {
      result = ((Type)expr).accept(this);
    } else if (expr instanceof Formula) {
      result = ((Formula)expr).accept(this);
    } else {
      assert (expr instanceof Term);
      result = ((Term)expr).accept(this);
    }
    return result;
  }
  
  /** {@inheritDoc} */
  @Override
  public Term caseIntConst(final BigInteger value) {
    return factory.createIntConst(value);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVariable(final String var) {
    return factory.createVariable(var);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseMkTuple(final Term... elements) {
    final List<Term> newElements = new ArrayList<Term>();
    for (final Term element : elements) {
      newElements.add(element.accept(this));
    }
    return factory.createMkTuple(newElements.toArray(new Term[]{}));
  }

  /** {@inheritDoc} */
  @Override
  public Term caseTupleAccess(final Term tuple, final int index) {
    final Term newTuple = tuple.accept(this);
    return factory.createTupleAccess(newTuple, index);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseUminus(final Term arg) {
    final Term newArg = arg.accept(this);
    return factory.createUminus(newArg);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseAddition(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createAddition(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseSubtraction(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createSubtraction(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseMultiplication(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createMultiplication(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseDivision(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createDivision(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseRemainder(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createRemainder(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term casePower(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createPower(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseLet(final String var,
                        final ValueType type,
                        final Term value,
                        final Term body) {
    final ValueType newType = (ValueType)type.accept(this);
    final Term newValue = value.accept(this);
    final Term newBody = body.accept(this);
    return factory.createLet(var, newType, newValue, newBody);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseLetProp(final String var,
                            final Formula value,
                            final Term body) {
    final Term newBody = body.accept(this);
    return factory.createLetProp(var, value, newBody);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseIfThenElse(final Formula cond,
                               final Term ifTrue,
                               final Term ifFalse) {
    final Formula newCond = cond.accept(this);
    final Term newIftrue = ifTrue.accept(this);
    final Term newIffalse = ifFalse.accept(this);
    return factory.createIfThenElse(newCond, newIftrue, newIffalse);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseArrayAccess(final Term arr, final Term index) {
    final Term newArr = arr.accept(this);
    final Term newIndex = index.accept(this);
    return factory.createArrayAccess(newArr, newIndex);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseArrayUpdate(final Term arr,
                                final Term index,
                                final Term value) {
    final Term newArr = arr.accept(this);
    final Term newIndex = index.accept(this);
    final Term newValue = value.accept(this);
    return factory.createArrayUpdate(newArr, newIndex, newValue);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseCall(final String func, final Term... args) {
    final List<Term> newArgs = new ArrayList<Term>();
    for (final Term arg : args) {
      newArgs.add(arg.accept(this));
    }
    return factory.createCall(func, newArgs.toArray(new Term[]{}));
  }

  /** {@inheritDoc} */
  @Override
  public Term caseConcat(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createConcat(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseSelectBits(final Term vec,
                               final int left,
                               final int right) {
    final Term newVec = vec.accept(this);
    return factory.createSelectBits(newVec, left, right);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseShl(final Term vec, final int offset) {
    final Term newVec = vec.accept(this);
    return factory.createShl(newVec, offset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseShr(final Term vec, final int offset) {
    final Term newVec = vec.accept(this);
    return factory.createShr(newVec, offset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseSX(final Term vec, final int offset) {
    final Term newVec = vec.accept(this);
    return factory.createSX(newVec, offset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseZeroExtend(final Term vec, final int offset) {
    final Term newVec = vec.accept(this);
    return factory.createZeroExtend(newVec, offset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseRepeat(final Term vec, final int count) {
    final Term newVec = vec.accept(this);
    return factory.createRepeat(newVec, count);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseRotateLeft(final Term vec, final int offset) {
    final Term newVec = vec.accept(this);
    return factory.createRotateLeft(newVec, offset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseRotateRight(final Term vec, final int offset) {
    final Term newVec = vec.accept(this);
    return factory.createRotateRight(newVec, offset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorAnd(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorAnd(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorOr(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorOr(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorXor(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorXor(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorNand(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorNand(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorNor(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorNor(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorXnor(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorXnor(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorCmp(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorCmp(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorSum(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorSum(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorDiff(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorDiff(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorUminus(final Term vec) {
    final Term newVec = vec.accept(this);
    return factory.createVectorUminus(newVec);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorProd(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorProd(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorShl(final Term vec, final Term offset) {
    final Term newVec = vec.accept(this);
    final Term newOffset = offset.accept(this);
    return factory.createVectorShl(newVec, newOffset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorLshr(final Term vec, final Term offset) {
    final Term newVec = vec.accept(this);
    final Term newOffset = offset.accept(this);
    return factory.createVectorLshr(newVec, newOffset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorRshr(final Term vec, final Term offset) {
    final Term newVec = vec.accept(this);
    final Term newOffset = offset.accept(this);
    return factory.createVectorRshr(newVec, newOffset);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorUdiv(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorUdiv(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorSdiv(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorSdiv(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorUrem(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorUrem(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorSrem(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorSrem(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorSmod(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorSmod(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorLt(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorLt(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorGt(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorGt(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorLe(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorLe(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorGe(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorGe(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorSlt(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorSlt(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorSgt(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorSgt(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorSle(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorSle(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Term caseVectorSge(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createVectorSge(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public StructuredType caseArray(final ValueType indexType,
                                  final ValueType valueType) {
    final ValueType newIndexType = (ValueType)indexType.accept(this);
    final ValueType newValueType = (ValueType)valueType.accept(this);
    return factory.createArray(newIndexType, newValueType);
  }

  /** {@inheritDoc} */
  @Override
  public StructuredType caseTuple(final ValueType... types) {
    final List<ValueType> newTypes = new ArrayList<ValueType>();
    for (final ValueType type : types) {
      newTypes.add((ValueType)type.accept(this));
    }
    return factory.createTuple(newTypes.toArray(new ValueType[]{}));
  }

  /** {@inheritDoc} */
  @Override
  public AtomicType caseRational() {
    return factory.createRational();
  }

  /** {@inheritDoc} */
  @Override
  public AtomicType caseInt() {
    return factory.createInt();
  }

  /** {@inheritDoc} */
  @Override
  public AtomicType caseBitvector(final int size) {
    return factory.createBitvector(size);
  }

  /** {@inheritDoc} */
  @Override
  public NonValueType caseProp() {
    return factory.createProp();
  }

  /** {@inheritDoc} */
  @Override
  public NonValueType caseFunction(final ValueType resultType,
                                   final ValueType... argTypes) {
    final ValueType newResultType = (ValueType)resultType.accept(this);
    final List<ValueType> newArgTypes = new ArrayList<ValueType>();
    for (final ValueType argType : argTypes) {
      newArgTypes.add((ValueType)argType.accept(this));
    }
    return factory.createFunction(newResultType, 
                                  newArgTypes.toArray(new ValueType[]{}));
  }

  /** {@inheritDoc} */
  @Override
  public NonValueType casePredicate(final ValueType... argTypes) {
    final List<ValueType> newArgTypes = new ArrayList<ValueType>();
    for (final ValueType argType : argTypes) {
      newArgTypes.add((ValueType)argType.accept(this));
    }
    return factory.createPredicate(newArgTypes.toArray(new ValueType[]{}));
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseTrue() {
    return factory.createTrue();
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseFalse() {
    return factory.createFalse();
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseFormulaVariable(final String name) {
    return factory.createFormulaVariable(name);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseNegation(final Formula arg) {
    final Formula newArg = arg.accept(this);
    return factory.createNegation(newArg);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseConjunction(final Formula left, final Formula right) {
    final Formula newLeft = left.accept(this);
    final Formula newRight = right.accept(this);
    return factory.createConjunction(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseDisjunction(final Formula left, final Formula right) {
    final Formula newLeft = left.accept(this);
    final Formula newRight = right.accept(this);
    return factory.createDisjunction(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseImplication(final Formula left, final Formula right) {
    final Formula newLeft = left.accept(this);
    final Formula newRight = right.accept(this);
    return factory.createImplication(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseXor(final Formula left, final Formula right) {
    final Formula newLeft = left.accept(this);
    final Formula newRight = right.accept(this);
    return factory.createXor(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseEquivalence(final Formula left, final Formula right) {
    final Formula newLeft = left.accept(this);
    final Formula newRight = right.accept(this);
    return factory.createEquivalence(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseEq(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createEq(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseNeq(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createNeq(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseDistinct(final Term... args) {
    final List<Term> newArgs = new ArrayList<Term>();
    for (final Term arg : args) {
      newArgs.add(arg.accept(this));
    }
    return factory.createDistinct(newArgs.toArray(new Term[]{}));
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseLt(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createLt(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseGt(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createGt(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseLe(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createLe(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseGe(final Term left, final Term right) {
    final Term newLeft = left.accept(this);
    final Term newRight = right.accept(this);
    return factory.createGe(newLeft, newRight);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseExists(final Iterable<VariableBinding> variables,
                            final Formula arg,
                            final Expression... patterns) {
    final List<VariableBinding> newVariables = 
        new ArrayList<VariableBinding>();
    for (final VariableBinding variable : variables) {
      final ValueType newVarType = 
          (ValueType)variable.getType().accept(this);
      final VariableBinding newVariable = 
          new VariableBinding(variable.getName(), newVarType);
      newVariables.add(newVariable);
    }
    final Formula newArg = arg.accept(this);
    final Expression[] newPatterns = 
        new Expression[patterns.length];
    for (int i = 0; i < patterns.length; i++) {
      newPatterns[i] = applyToExpression(patterns[i]);
    }
    
    return factory.createExists(newVariables, newArg, newPatterns);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseForall(final Iterable<VariableBinding> variables,
                            final Formula arg,
                            final Expression... patterns) {
    final List<VariableBinding> newVariables = 
        new ArrayList<VariableBinding>();
    for (final VariableBinding variable : variables) {
      final ValueType newVarType = 
          (ValueType)variable.getType().accept(this);
      final VariableBinding newVariable = 
          new VariableBinding(variable.getName(), newVarType);
      newVariables.add(newVariable);
    }
    final Formula newArg = arg.accept(this);
    final Expression[] newPatterns = 
        new Expression[patterns.length];
    for (int i = 0; i < patterns.length; i++) {
      newPatterns[i] = applyToExpression(patterns[i]);
    }
    return factory.createForall(newVariables, newArg, newPatterns);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseFLet(final String var,
                          final ValueType type,
                          final Term value,
                          final Formula body) {
    final ValueType newType = (ValueType)type.accept(this);
    final Term newValue = value.accept(this);
    final Formula newBody = body.accept(this);
    return factory.createFLet(var, newType, newValue, newBody);
  }

  /** {@inheritDoc} */
  @Override
  public Formula caseFLetProp(final String var,
                                final Formula value,
                                final Formula body) {
    final Formula newValue = value.accept(this);
    final Formula newBody = body.accept(this);
    return factory.createFLetProp(var, newValue, newBody);
  }

  @Override
  public Formula casePredicateCall(final String pred, final Term... args) {
    final List<Term> newArgs = new ArrayList<Term>();
    for (final Term arg : args) {
      newArgs.add(arg.accept(this));
    }
    return factory.createPredicateCall(pred, newArgs.toArray(new Term[]{}));
  }
}
