Require Export ZArith ZArith.Int.
Local Open Scope Z_scope.

(* Array type notation *)
Notation "'array' [ 'Z' ] 'of' Y" := (Z -> Y) (at level 35, right associativity,
                                            format "'array' [ 'Z' ]  'of'  Y").

(* Array lookup *)

Definition lookup {Y} (A : array[Z] of Y) i := (A i).

Notation "A [ i ]" := (lookup A i) (at level 20, left associativity).

(* Array update *)

Definition update {Y} (A : array[Z] of Y) (where_ : Z) (val : Y) :=
		      fun (where__ : Z) =>  if (Z.eqb where_ where__) then val else A[where__].

Notation "A [ i <- v ]" := (update A i v) (at level 20, left associativity). 

Lemma update_elsewhere: forall Y A i (v:Y) j, i <> j -> A[i <- v][j] = A[j].
Proof.
  intros.
  unfold update, lookup.
  apply Z.eqb_neq in H.
  now rewrite H.
Qed.

Lemma update_here: forall Y A i (v:Y) j, i = j -> A[i <- v][j] = v.
Proof.
  intros.
  unfold update, lookup.
  apply Z.eqb_eq in H.
  now rewrite H.
Qed.




(* Tactics *)

(* simpifies ((update A i v) [j]) everywhere *)

Ltac solve_this_update A i v j :=
  let H := fresh "H" in
  idtac A; idtac i; idtac j;
  first [ rewrite !(update_elsewhere A i v j) in * |- * by auto 1 with zarith; idtac "<> poszło!"
        | rewrite !(update_here A i v j) in * |- * by auto 1 with zarith; idtac "= poszło!"
        | (assert (i<>j) as H by omega; idtac "assert i<>j poszedl";
            rewrite !(update_elsewhere A i v j) in * |- * by exact H)
        | (assert (i=j) as H by omega; idtac "assert i=j poszedl";
            rewrite !(update_here A i v j) in * |- * by exact H)
        ].

(* two cases i=j and i<>j *)

Ltac split_this_update A i v j := idtac "split"; destruct (Z.eq_dec i j).


(* find "update" and call tac on it *)

Ltac find_update tac :=
  match goal with 
  | [ |- context [ (update ?A ?i ?v)[?j] ] ] => 
    tac A i v j ; idtac "w G"
  | [ hyp : context [ (update ?A ?i ?v)[?j] ] |- _ ] => 
    tac A i v j ; idtac "w HYP"
  end.


Ltac ss_up :=
  repeat
    (repeat find_update solve_this_update; (* solve all easy updates *)
     auto with zarith;                     (* clean up *)
     try find_update split_this_update).   (* try to split one "uneasy" update (if any) *)
     (* after splitting an "uneasy" update becomes easy in both branches *)
     (* that is the primary reason for the external repeat *)


(* nie wolno w nich niestety używać idtac "a kuku" *)

Ltac is_pos_const t := 
  match t with 
    | xH => true
    | xO ?x => is_pos_const x
    | xI ?x => is_pos_const x
    | _ => false
  end.

Ltac is_const t :=
  match t with
    | Z0 => true
    | Zpos ?t => is_pos_const t
    | Zneg ?t => is_pos_const t
    | _ => false
  end.

Ltac check_const t :=
  match is_const t with
  | true => idtac t; idtac "stała!"
  | false => idtac t; idtac "niestała; fail"; fail 1
  end.

Ltac rewconst :=
  match goal with
    | [H: ?a = ?b |- _ ] => check_const b; idtac "robimy rewrite ->"; idtac a; rewrite -> H in *; idtac " "; idtac " "
    | [H: ?a = ?b |- _ ] => check_const a; idtac "robimy rewrite <-"; idtac b; rewrite <- H in *; idtac " "; idtac " "
  end.

(*
Tests: 
* )

Goal forall n m, 4=n -> m=5 -> (n + 1) - 1 = m - 2 + 1.
intros.
ring_simplify.
rewconst.
rewconst.
ring_simplify.
trivial.
Qed.
*)

Ltac instconst :=
  match goal with
    | [H: ?a = ?b |- _ ] => check_const b; idtac "robimy rewrite ->"; idtac a; rewrite -> H in *; idtac " "
    | [H: ?a = ?b |- _ ] => check_const a; idtac "robimy rewrite <-"; idtac b; rewrite <- H in *; idtac " "
    | [H: ?a = ?b |- _ ] => idtac "robimy rewrite w ciemno!"; rewrite H in *
  end
  ||
  ring_simplify.


Ltac instconstn :=
  match goal with
    | [H: ?a = ?b |- _ ] => check_const b; idtac "robimy rewrite ->"; idtac a; rewrite -> H in *; idtac " "
    | [H: ?a = ?b |- _ ] => check_const a; idtac "robimy rewrite <-"; idtac b; rewrite <- H in *; idtac " "
    | [H: ?a = ?b |- _ ] => idtac "robimy rewrite w ciemno!"; rewrite H in *
    | [H: ?a = ?b |- _ ] => idtac "robimy rewrite w ciemno!"; rewrite H in *
  end
  ||
  ring_simplify.


Ltac size t :=
  match is_const t with
  | true => constr:(-1)
  | false => 
      match t with
      | ?X ?Y => 
          let xs := size X in 
          let ys := size Y in 
          let xss := constr:xs in
          let yss := constr:ys in
          constr:(xs + ys)
      | _ => constr:(2)
      end
  end.

Ltac ssize t :=
  match t with
    | Z0 => constr:(-1)
    | ?X ?Y => size t
    | _ => constr:(100)
  end.

Definition ID (E:Prop) := E.

Ltac rew_to_smaller := 
  match goal with
    | [H: ?a = ?b |- _ ] => idtac "eq"; 
      let asss := ssize a in (* *) idtac asss; (* *)
      let ass := eval simpl in asss in (* *) idtac ass; (* *) 
      let bsss := ssize b in (* idtac bsss; *)
      let bss := eval simpl in bsss in (* *) idtac bss; (* *)
      let cmp := eval compute in (Z.compare ass bss) in (* *) idtac cmp; (* *)
      match cmp with 
      | Lt => rewrite <- H in *; change (ID (a=b)) in H
      | Gt => rewrite -> H in *; change (ID (a=b)) in H
      | Eq => rewrite -> H in *; change (ID (a=b)) in H
      end
  end.

Ltac zawin H :=
  let E := type of H in
  change (ID E) in H.

Ltac rseq :=
  match goal with
    | [H: ?a = ?b |- _ ] => ring_simplify in H; zawin H
  end.

Ltac uprosc := repeat rew_to_smaller; unfold ID in *; repeat rseq; unfold ID in *; ring_simplify.

(* Tests * )
Goal forall n m a, 1- m + (2-a + n)=3 -> m = 2*(n-1) + 3 -> (2*(m-1) + 3) * (n + 1) = 16.
intros.
uprosc.


Abort.

( *
ring_simplify in H.
rew_to_smaller.
ring_simplify.

Undo.
symmetry in H0.
rew_to_smaller.

rew_to_smaller. ( * tu się rewrite H0 nie powiódł i zbacktrackował do H :) * )
trivial.
Qed.
* )

Goal forall n m, n = 0 -> 1 = m -> m+1 = 2+n.
intros.
rew_to_smaller.
rew_to_smaller.
trivial.
Qed.

*)



Ltac haha_solve :=
  repeat autounfold with haha_predicates;
  intuition;
  try (timeout 10 (do 5 uprosc));
  try (timeout 10 ss_up);
  try (timeout 10 firstorder);
  try omega.
