package net.afyre.haha.demo.util.symboltable.tests;

import static org.junit.Assert.*;

import org.junit.Test;

import net.afyre.haha.demo.util.symboltable.Frame;

/**
 * Common tests for all {@link Frame} implementations.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public abstract class FrameTest {
  /**
   * Create an empty, root frame.
   * 
   * For simplicity we use Strings as symbols and symbol information.
   * 
   * @return New frame.
   */
  protected abstract Frame<String, String> createRootFrame();

  /** Root frame has no parent. */
  @Test
  public void testRootParentIsNull() {
    final Frame<String, String> root = createRootFrame();
    assertNull("Root frame has a parent", root.getParent());
  }

  /** Child frame returns the correct parent. */
  @Test
  public void testChildParent() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    assertSame("Wrong parent", root, child.getParent());
  }

  /** Grandchild frame returns the correct parent. */
  @Test
  public void testGrandchildParent() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    final Frame<String, String> grandchild = child.createChildFrame();
    assertSame("Wrong parent", child, grandchild.getParent());
  }
  
  /** Test symbol not present in new root frame. */
  @Test
  public void testNoSymbolInEmptyRootFrame() {
    final Frame<String, String> root = createRootFrame();
    
    assertNull("New root frame is not empty", root.find("TEST"));
  }
  
  /** Attempt to set information for null symbol causes an exception. */
  @Test(expected=NullPointerException.class)
  public void testSetSymbolInfoNullSymbol() {
    final Frame<String, String> root = createRootFrame();
    root.setSymbolInfo(null, "TESTINFO");
  }
  
  /** Attempt to set null information for a symbol causes an exception. */
  @Test(expected=NullPointerException.class)
  public void testSetSymbolInfoNullInfo() {
    final Frame<String, String> root = createRootFrame();
    root.setSymbolInfo("TEST", null);
  }
  
  /** Lookup of a null symbol causes an exception. */
  @Test(expected=NullPointerException.class)
  public void testFindNull() {
    final Frame<String, String> root = createRootFrame();
    root.find(null);
  }
  
  /** Symbol added to an empty root frame can be found. */
  @Test
  public void testSetSymbolInfoResultNotNull() {
    final Frame<String, String> root = createRootFrame();
    root.setSymbolInfo("TEST", "TESTINFO");
    final String result = root.find("TEST");
    assertNotNull("Test symbol not inserted", result);
  }

  /** Correct information is returned for newly inserted symbol. */
  @Test
  public void testSetSymbolInfoResultValid() {
    final Frame<String, String> root = createRootFrame();
    root.setSymbolInfo("TEST", "TESTINFO");
    final String result = root.find("TEST");
    assertEquals("Wrong symbol information returned", "TESTINFO", result);
  }
  
  /** Returned child frame is not null. */
  @Test
  public void testCreateChildFrameNotNull() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    assertNotNull("Child frame is null", child);
  }
  
  /** New child frame is empty. */
  @Test
  public void testChildFrameIsEmpty() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    assertNull("New child frame is not empty", child.find("TEST"));
  }
  
  /** Symbol added to an empty child frame can be found. */
  @Test
  public void testSetSymbolInfoInChildResultNotNull() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    child.setSymbolInfo("TEST", "TESTINFO");
    final String result = child.find("TEST");
    
    assertNotNull("Test symbol not inserted", result);
  }

  /** Correct information is returned for new symbol in child frame. */
  @Test
  public void testSetSymbolInfoInChildResultValid() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    child.setSymbolInfo("TEST", "TESTINFO");
    final String result = child.find("TEST");
    
    assertEquals("Wrong symbol information returned", "TESTINFO", result);
  }
  
  /** Symbol added to a child frame is not present in parent. */
  @Test
  public void testSetSymbolInfoNotAddedToParent() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    child.setSymbolInfo("TEST", "TESTINFO");
    final String result = root.find("TEST");
    
    assertNull("Root frame was modified", result);
  }
  
  /** Symbol from the root frame can be found in child frame. */
  @Test
  public void testSetSymbolInfoInRootFindInChild1() {
    final Frame<String, String> root = createRootFrame();
    root.setSymbolInfo("TEST", "TESTINFO");
    final Frame<String, String> child = root.createChildFrame();
    
    final String result = child.find("TEST");
    
    assertNotNull("Test symbol not inserted", result);    
  }
  
  /** Information in child frame overrides that from the root frame. */
  @Test
  public void testSymbolShadowing() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    root.setSymbolInfo("TEST", "ROOTTESTINFO");
    child.setSymbolInfo("TEST", "CHILDTESTINFO");
    final String result = child.find("TEST");
    
    assertEquals("Wrong symbol information returned", 
                 "CHILDTESTINFO", result);    
  }
  
  /** Different symbols do not collide. */
  @Test
  public void testTwoSymbols() {
    final Frame<String, String> root = createRootFrame();

    root.setSymbolInfo("TEST1", "TESTINFO1");
    root.setSymbolInfo("TEST2", "TESTINFO2");
    
    assertEquals("Wrong information for firat symbol", 
                 "TESTINFO1", root.find("TEST1"));
    assertEquals("Wrong information for second symbol", 
                 "TESTINFO2", root.find("TEST2"));
  }
  
  /** No shadowing between different symbols. */
  @Test
  public void testTwoSymbols2() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    root.setSymbolInfo("TEST1", "TESTINFO1");
    child.setSymbolInfo("TEST2", "TESTINFO2");

    assertEquals("Wrong information for firat symbol", 
                 "TESTINFO1", child.find("TEST1"));
    assertEquals("Wrong information for second symbol", 
                 "TESTINFO2", child.find("TEST2"));
  }
  
  /** Symbol added to the root frame can be found in child frame. */
  @Test
  public void testSetSymbolInfoInRootFindInChild2() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    root.setSymbolInfo("TEST", "TESTINFO");
    final String result = child.find("TEST");
    
    assertNotNull("Test symbol not inserted", result);    
  }
  
  /** Symbol information is correctly updated in the root frame. */
  @Test
  public void testUpdateInRoot() {
    final Frame<String, String> root = createRootFrame();
    
    root.setSymbolInfo("TEST", "OLDTESTINFO");
    root.setSymbolInfo("TEST", "NEWTESTINFO");
    
    final String result = root.find("TEST");
    
    assertNotNull("Symbol not inserted (or removed)", result);
    assertFalse("Symbol information not updated", 
                "OLDTESTINFO".equals(result));
    assertEquals("Wrong symbol information returned", "NEWTESTINFO", result);    
  }
  
  /** Information update is visible in a child frame. */
  @Test
  public void testUpdateInRootFindInChild() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    root.setSymbolInfo("TEST", "OLDTESTINFO");
    root.setSymbolInfo("TEST", "NEWTESTINFO");
    
    final String result = child.find("TEST");
    
    assertNotNull("Symbol not inserted (or removed)", result);
    assertFalse("Symbol information not updated", 
                "OLDTESTINFO".equals(result));
    assertEquals("Wrong symbol information returned", "NEWTESTINFO", result);    
  }  
  
  /** Update in child does not affect the parent. */
  @Test
  public void testUpdateInChildFindInRoot() {
    final Frame<String, String> root = createRootFrame();
    final Frame<String, String> child = root.createChildFrame();
    
    root.setSymbolInfo("TEST", "ROOTTESTINFO");
    child.setSymbolInfo("TEST", "CHILDTESTINFO1");
    child.setSymbolInfo("TEST", "CHILDTESTINFO2");

    final String result = root.find("TEST");
    
    assertEquals("Root frame modified", "ROOTTESTINFO", result);
  }
}
