package net.afyre.haha.demo.util.tests;

import static org.junit.Assert.*;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import org.jmock.Expectations;
import org.jmock.Mockery;
import org.jmock.integration.junit4.JUnit4Mockery;
import org.jmock.lib.legacy.ClassImposteriser;
import org.junit.Before;
import org.junit.Test;

import net.afyre.haha.demo.util.TeeOutputStream;

/**
 * Test for the {@link TeeOutputStream} class.
 *  
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class TeeOutputStreamTest {
  /** JMock context. */
  Mockery mockery;
  
  /** Create JMock context. */
  @Before
  public void setUp() {
    mockery = new JUnit4Mockery() {{
      setImposteriser(ClassImposteriser.INSTANCE);
    }};
  }
  
  /** Exception is thrown when a null array is passed to getInstance. */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceNullArray1() {
    TeeOutputStream.getInstance((OutputStream[])null);
  }
  
  /** Exception is thrown when a null parameter is passed to getInstance. */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceNull1() {
    TeeOutputStream.getInstance((OutputStream)null);
  }
  
  /** Exception is thrown when a null parameter is passed to getInstance. */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceNull2() {
    TeeOutputStream.getInstance(new ByteArrayOutputStream(), null);
  }

  /** Exception is thrown when a null array is passed to getInstance. */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceNullArray2() {
    TeeOutputStream.getInstance(true, (OutputStream[])null);
  }
  
  /** Exception is thrown when a null parameter is passed to getInstance. */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceNull3() {
    TeeOutputStream.getInstance(true, (OutputStream)null);
  }
  
  /** Exception is thrown when a null parameter is passed to getInstance. */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceNull4() {
    TeeOutputStream.getInstance(true, new ByteArrayOutputStream(), null);
  }
  
  /** It is possible to create an empty instance. */
  @Test
  public void testNoOutputs() {
    try {
      TeeOutputStream.getInstance();
      TeeOutputStream.getInstance(true);
    } catch (final Exception e) {
      fail("Unable to create an instance with no outputs");
    }
  }
  
  /** Flush works for streams without output. */
  @Test
  public void testFlushEmptyStream() {
    final TeeOutputStream stream = TeeOutputStream.getInstance();
    try {
      stream.flush();
    } catch (final IOException e) {
      fail("Unable to flush a stream with no outputs.");
    }
  }
  
  /** All streams are flushed when flush() is called. */
  @Test
  public void testFlush() throws IOException {
    final OutputStream[] mocks = createMocks();
    
    // Setup expectations.
    for (final OutputStream mock : mocks) {
      mockery.checking(new Expectations() {{
        oneOf(mock).flush();
      }});
    }
    
    // Execute test
    final TeeOutputStream stream = TeeOutputStream.getInstance(mocks);
    stream.flush();
  }
  
  /** Close works for streams without output. */
  @Test
  public void testCloseEmptyStream() {
    final TeeOutputStream stream = TeeOutputStream.getInstance();
    try {
      stream.close();
    } catch (final IOException e) {
      fail("Unable to close a stream with no outputs.");
    }
  }
  
  /** All output streams are closed on close() call by default. */
  @Test
  public void testCloseDefault() throws IOException {
    final OutputStream[] mocks = createMocks();
    
    // Setup expectations.
    for (final OutputStream mock : mocks) {
      mockery.checking(new Expectations() {{
        oneOf(mock).close();
      }});
    }
    
    // Execute test
    final TeeOutputStream stream = TeeOutputStream.getInstance(mocks);
    stream.close();    
  }
  
  /** All output streams are closed on close() call if requested. */
  @Test
  public void testCloseExplicitTrue() throws IOException {
    final OutputStream[] mocks = createMocks();
    
    // Setup expectations.
    for (final OutputStream mock : mocks) {
      mockery.checking(new Expectations() {{
        oneOf(mock).close();
      }});
    }
    
    // Execute test
    final TeeOutputStream stream = TeeOutputStream.getInstance(true, mocks);
    stream.close();    
  }  
  
  /** Output streams are not closed on close() call when requested. */
  @Test
  public void testCloseFalse() throws IOException {
    final OutputStream[] mocks = createMocks();
    
    // Setup expectations - no calls expected.
    
    // Execute test
    final TeeOutputStream stream = TeeOutputStream.getInstance(false, mocks);
    stream.close();    
  }
  
  /** Write(int) is passed to all output streams. */
  @Test
  public void testWriteInt() throws IOException {
    final OutputStream[] mocks = createMocks();
    
    // Setup expectations.
    for (final OutputStream mock : mocks) {
      mockery.checking(new Expectations() {{
        oneOf(mock).write(42);
      }});
    }
    
    // Execute test
    final TeeOutputStream stream = TeeOutputStream.getInstance(true, mocks);
    stream.write(42);    
  }  

  /** Write(byte[], int, int) is passed to all output streams. */
  @Test
  public void testWriteByteWithOffset() throws IOException {
    final OutputStream[] mocks = createMocks();
    final byte[] data = { 100, 117, 112, 97 };
    
    // Setup expectations.
    for (final OutputStream mock : mocks) {
      mockery.checking(new Expectations() {{
        oneOf(mock).write(data, 2, 1);
      }});
    }
    
    // Execute test
    final TeeOutputStream stream = TeeOutputStream.getInstance(true, mocks);
    stream.write(data, 2, 1);    
  }  
  
  /** Write(byte[]) is passed to all output streams. */
  @Test
  public void testWriteByte() throws IOException {
    final OutputStream[] mocks = createMocks();
    final byte[] data = { 100, 117, 112, 97 };
    
    // Setup expectations.
    for (final OutputStream mock : mocks) {
      mockery.checking(new Expectations() {{
        oneOf(mock).write(data);
      }});
    }
    
    // Execute test
    final TeeOutputStream stream = TeeOutputStream.getInstance(true, mocks);
    stream.write(data);    
  }  
  
  /** Create a few mocked output streams. */
  private OutputStream[] createMocks() {
    // Chosen by fait dice roll. Guaranteed to be random.
    final int count = 4;
    
    final OutputStream[] mocks = new OutputStream[4];
    for (int i = 0; i < count; i++) {
      final OutputStream mock = 
          mockery.mock(OutputStream.class, "mocks[" + i +"]");
      mocks[i] = mock;
    }
    
    return mocks;
  }
}
