package net.afyre.haha.demo.util;

/**
 * Objects used to calculate FNV hashes.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class FNVBuilder implements HashBuilder {
  /** Prime used in 32-bit FNV hashing. */
  private static final int FNV_PRIME = 16777619;
  
  /** Initial value used in 32-bit FNV hashing. */
  private static final int FNV_OFFSET = 0x811C9DC5;
  
  /** Octet size in bits. */
  private static final int OCTET_SIZE = 8;
  
  /** Current hash value. */
  private int hash;
  
  /** Constructor. */
  private FNVBuilder() {
    hash = FNV_OFFSET;
  }
  
  /**
   * Get a new instance of this class.
   * @return Instance.
   */
  public static FNVBuilder getInstance() {
    return new FNVBuilder();
  }
  
  /**
   * Get current hash value.
   * @return Hash value of current octet sequence.
   */
  public int getHash() {
    return hash;
  }
  
  /**
   * Append a byte to the hashed sequence.
   * @param b Byte value.
   */
  public void appendByte(final byte b) {
    hash = hash ^ b;
    hash = hash * FNV_PRIME;
  }
  
  /**
   * Append an integer to the hashed sequence.
   * Big-endian ordering is used when encoding 
   * the integer as a sequence of octets.
   * @param i Integer value.
   */
  public void appendInt(final int i) {
    int offset = 0;

    hash = hash ^ ((i >>> offset) & 0xff);
    hash = hash * FNV_PRIME;
    offset += OCTET_SIZE;
    
    hash = hash ^ ((i >>> offset) & 0xff);
    hash = hash * FNV_PRIME;
    offset += OCTET_SIZE;
    
    hash = hash ^ ((i >>> offset) & 0xff);
    hash = hash * FNV_PRIME;
    offset += OCTET_SIZE;
    
    hash = hash ^ ((i >>> offset) & 0xff);
    hash = hash * FNV_PRIME;
  }
    
  /**
   * Clear the hashed sequence.
   */
  public void reset() {
    hash = FNV_OFFSET;
  }
}
