package net.afyre.haha.demo.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * A wrapper for input streams which copies 
 * all bytes read to a given output stream.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class TeeInputStream extends InputStream {
  /** Original input stream. */
  private final InputStream input;
  
  /** Output stream. */
  private final OutputStream output;
  
  /** Should the output stream be closed by this wrapper? */
  private final boolean closeOutput;
  
  /**
   * Create a wrapper for a given input stream.
   * @param input Original input stream.
   * @param output Output stream.
   * @param closeOutput Should the input stream be closed by 
   *                    the returned wrapper?
   * @return Wrapped stream.
   * @throws NullPointerException If given a null argument.
   */
  public static TeeInputStream getInstance(final InputStream input, 
                         final OutputStream output, 
                         final boolean closeOutput) {
    return new TeeInputStream(input, output, closeOutput);
  }

  /**
   * Constructor.
   * @param input Original input stream.
   * @param output Output stream.
   * @param closeOutput Should the input stream be closed by this wrapper?
   * @throws NullPointerException If given a null argument.
   */
  private TeeInputStream(final InputStream input, 
                         final OutputStream output, 
                         final boolean closeOutput) {
    if (input == null) {
      throw new NullPointerException("input");
    } else if (output == null) {
      throw new NullPointerException("output");
    }
    this.input = input;
    this.output = output;
    this.closeOutput = closeOutput;
  }

  /** @{inheritDoc} */
  @Override
  public void close() throws IOException {
    input.close();
    if (closeOutput) {
      output.close();
    }
  }
  
  /** @{inheritDoc} */
  @Override
  public int read() throws IOException {
    final int result = input.read();
    if (result != -1) {
      output.write(result);
    }
    return result;
  }
  
  /** @{inheritDoc} */
  @Override
  public int read(final byte[] buffer, 
                  final int offset, 
                  final int length) throws IOException {
    final int result = input.read(buffer, offset, length);
    if (result != -1) {
      output.write(buffer, offset, result);
    }
    return result;    
  }
  
  /** @{inheritDoc} */
  @Override
  public int read(final byte[] buffer) throws IOException {
    final int result = input.read(buffer);
    if (result != -1) {
      output.write(buffer);
    }
    return result;    
  }
}
