package net.afyre.haha.demo.types;

import java.util.WeakHashMap;

import net.afyre.haha.demo.haha.Axiom;
import net.afyre.haha.demo.haha.Expression;
import net.afyre.haha.demo.haha.HahaPackage.Literals;
import net.afyre.haha.demo.haha.Statement;
import net.afyre.haha.demo.haha.util.HahaSwitch;

import org.eclipse.emf.ecore.EObject;

/**
 * The typechecker maintains a mapping from expressions to types.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class TypeChecker {
  private final WeakHashMap<EObject, ValueType> types;
  private final WeakHashMap<EObject, TypeError> errors;
  private final TypeUtil util;
  
  public TypeChecker() {
    types = new WeakHashMap<EObject, ValueType>();
    errors = new WeakHashMap<EObject, TypeError>();
    util = new TypeUtil();
  }

  /**
   * Update the mapping by scanning given program element.
   * @param obj Program element to be typechecked.
   */
  public void check(final EObject obj) {
    new Switch().doSwitch(obj);
  }
  
  /**
   * Return stored type for given expression.
   * 
   * Return {@code null} if
   *   - The expression was not typechecked ({@link check()}), or
   *   - There is a type error in the expression ({@link getError()})
   * @param expr Expression.
   * @return Type or {@code null}.
   */
  public ValueType getType(final EObject expr) {
    return types.get(expr);
  }
  
  /**
   * Get error information associated with an expression.
   * Note that errors in subexpressions will not be returned.
   * @param expr Expression.
   * @return Error information or {@code null} (no error).
   */
  public TypeError getError(final EObject expr) {
    return errors.get(expr);
  }
  
  /** Switch used to deconstruct program elements. */
  private final class Switch extends HahaSwitch<ValueType> {
    // Fail fast for unsupported (most likely forgotten) types.
    
    @Override
    public ValueType defaultCase(final EObject obj) {
      throw new RuntimeException("Unsupported language construct: " + 
                                 obj.eClass().getName());
    }
    
    @Override
    public ValueType caseStatement(final Statement stmt) {
      throw new RuntimeException("Unsupported statement: " + 
                                 stmt.eClass().getName());
    }

    @Override
    public ValueType caseExpression(final Expression expr) {
      throw new RuntimeException("Unsupported expression type: " + 
                                 expr.eClass().getName());
    }
    
    @Override
    public ValueType caseAxiom(final Axiom axiom) {
      final ValueType type = doSwitch(axiom.getFormula());
      if (type != null && !util.isBool(type)) {
        errors.put(axiom.getFormula(), 
                   new TypeError("Wrong type: Boolean expected, " + 
                                    util.getTypeName(type) + " found",
                                 axiom, Literals.AXIOM__FORMULA, -1));
      }
      return type;
    }
    
    // TODO: Implement.
  }
}
