package net.afyre.haha.demo.types;

/**
 * Utility methods for working with types.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class TypeUtil {
  /**
   * Check if a given type represents fixed-width integers.
   * @param type Type to check.
   * @return True for type INT, false otherwise.
   */
  public boolean isInteger(final Type type) {
    return type.doSwitch(new FalseTypeSwitch() {
       @Override public Boolean caseInteger() { return true; }
    });
  }

  /**
   * Check if a given type represents unbounded integers.
   * @param type Type to check.
   * @return True for type Z, false otherwise.
   */
  public boolean isZ(final Type type) {
    return type.doSwitch(new FalseTypeSwitch() {
       @Override public Boolean caseZ() { return true; }
    });
  }
  
  /**
   * Check if a given type represents booleans.
   * @param type Type to check.
   * @return True for type BOOL, false otherwise.
   */
  public boolean isBool(final Type type) {
    return type.doSwitch(new FalseTypeSwitch() {
       @Override public Boolean caseBoolean() { return true; }
    });
  }
  
  /**
   * Get human-readable name of a type.
   * @param type Type.
   * @return Type name.
   */
  public String getTypeName(final Type type) {
    return type.doSwitch(new TypeSwitch<String>() {

      @Override
      public String caseArray(final ValueType elementType) {
        return "array[" + elementType.doSwitch(this) + "]";
      }

      @Override
      public String caseZ() {
        return "Z";
      }

      @Override
      public String caseInteger() {
        return "Int";
      }

      @Override
      public String caseBoolean() {
        return "Boolean";
      }

      @Override
      public String caseFunction(final ValueType resultType, 
                                 final ValueType... argTypes) {
        final StringBuilder result = new StringBuilder();
        for (final ValueType argType : argTypes) {
          result.append(argType.doSwitch(this));
          result.append(" -> ");
        }
        result.append(resultType.doSwitch(this));
        return result.toString();
      }
    });
  }
}
