package net.afyre.haha.demo.validation;

import java.util.Iterator;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.validation.Check;

import com.google.common.collect.Iterators;

import net.afyre.haha.demo.haha.ActualArgs;
import net.afyre.haha.demo.haha.Argument;
import net.afyre.haha.demo.haha.ArgumentList;
import net.afyre.haha.demo.haha.ArrayAccess;
import net.afyre.haha.demo.haha.ArrayType;
import net.afyre.haha.demo.haha.Assertion;
import net.afyre.haha.demo.haha.Assignment;
import net.afyre.haha.demo.haha.Axiom;
import net.afyre.haha.demo.haha.BinaryArithOp;
import net.afyre.haha.demo.haha.BinaryBoolOp;
import net.afyre.haha.demo.haha.Block;
import net.afyre.haha.demo.haha.BoolLiteral;
import net.afyre.haha.demo.haha.Call;
import net.afyre.haha.demo.haha.Callable;
import net.afyre.haha.demo.haha.ComparisonOp;
import net.afyre.haha.demo.haha.Cond;
import net.afyre.haha.demo.haha.CounterVariable;
import net.afyre.haha.demo.haha.Exists;
import net.afyre.haha.demo.haha.Expression;
import net.afyre.haha.demo.haha.Forall;
import net.afyre.haha.demo.haha.Function;
import net.afyre.haha.demo.haha.HahaPackage;
import net.afyre.haha.demo.haha.HoareTriple;
import net.afyre.haha.demo.haha.IntLiteral;
import net.afyre.haha.demo.haha.LValue;
import net.afyre.haha.demo.haha.Loop;
import net.afyre.haha.demo.haha.LoopCounter;
import net.afyre.haha.demo.haha.LoopInvariant;
import net.afyre.haha.demo.haha.Postcondition;
import net.afyre.haha.demo.haha.Precondition;
import net.afyre.haha.demo.haha.Predicate;
import net.afyre.haha.demo.haha.PrimitiveType;
import net.afyre.haha.demo.haha.QuantifiedVar;
import net.afyre.haha.demo.haha.Quantifier;
import net.afyre.haha.demo.haha.ResultVar;
import net.afyre.haha.demo.haha.SimpleType;
import net.afyre.haha.demo.haha.Skip;
import net.afyre.haha.demo.haha.Statement;
import net.afyre.haha.demo.haha.Type;
import net.afyre.haha.demo.haha.TypeUtils;
import net.afyre.haha.demo.haha.UnaryArithOp;
import net.afyre.haha.demo.haha.UnaryBoolOp;
import net.afyre.haha.demo.haha.Var;
import net.afyre.haha.demo.haha.Variable;
import net.afyre.haha.demo.haha.util.HahaSwitch;
import static net.afyre.haha.demo.haha.HahaPackage.Literals;

/**
 * The class checks the type correctness of the expressions and instructions
 * in the HAHA language.
 * 
 * @author Tadeusz Sznuk 
 * @author Aleksy Schubert
 */
public class HahaJavaValidator extends AbstractHahaJavaValidator {
  private final TypeUtils typeUtils = TypeUtils.getInstance();
  private final ResultChecker resultChecker = new ResultChecker();
  
  ////////////////////////////////////////////////////////////////////////////
  //// Preconditions
  ////////////////////////////////////////////////////////////////////////////
  /**
   * Consistency check for functions verifies that there is no reference
   * to the result of the function in all its preconditions.
   * 
   * @param fun the function to check
   */
  @Check
  public void checkPreconditionsInFunction(final Function fun) {
    EList<Precondition> pconds = fun.getPreconditions();
    ResultVar rvar = fun.getResultVar();
    String rname = rvar.getFunction().getName();
    for (Precondition precondition : pconds) {
      checkResultInPrecondition(precondition, rname);
    }
  }
  
  /**
   * Checks if given function name occurs in the precondition.
   * 
   * @param pre the precondition to check the occurrence of the variable
   * @param rname the variable name
   */
  private void checkResultInPrecondition(final Precondition pre, String rname) {
    final Iterator<EObject> it = pre.eAllContents();
    while (it.hasNext()) {
      final EObject obj = (EObject)it.next();
      if (obj instanceof Var) {
        final Var var = (Var)obj;
        if (var.getName().equals(rname)) {
          error("Cannot refer to function result in a preconditon.", 
                var, Literals.VAR__SYMBOL, -1);
        }
      }
    }
  }

  ////////////////////////////////////////////////////////////////////////////
  //// Result assignment
  ////////////////////////////////////////////////////////////////////////////
  /**
   * Check for unassigned function result.
   * 
   * @param func Function to be checked.
   */
  @Check
  public void checkResultIsAssigned(final Function func) {
    if (!resultChecker.doSwitch(func.getBody())) {
      warning("Function result not assigned.", func, Literals.CALLABLE__NAME, -1);
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////
  //// Quantifiers
  ////////////////////////////////////////////////////////////////////////////
  /**
   * Consistency check for quantified expressions. We allow only quantification
   * over simple types, i.e. INT, Z, BOOLEAN.
   * 
   * @param expr the quantification expression to check
   */
  @Check
  public void checkForallType(final Quantifier expr) {
    for (final QuantifiedVar variable : expr.getVars()) {
      if (! (variable.getType() instanceof SimpleType)) {
        error("Quantification over this type is not allowed", 
              variable, Literals.DECLARED_VARIABLE__TYPE, -1);
      }      
    }
  }

  ////////////////////////////////////////////////////////////////////////////
  //// Restrictions on program expressions
  ////////////////////////////////////////////////////////////////////////////
  /**
   * Conditionals have program expressions in their conditions so restriction
   * on them must be checked.
   * 
   * @param cond the conditional instruction to check the restrictions for its
   *   condition
   */
  @Check
  public void checkProgramRestrictionsForIfConditions(final Cond cond) {
    checkRestrictionsForProgramExpr(cond.getCondition());
  }

  /**
   * Loops have program expressions in their conditions so restriction
   * on them must be checked.
   * 
   * @param loop the loop instruction to check the restrictions for its
   *   condition
   */
  @Check
  public void checkProgramRestrictionsForLoopConditions(final Loop loop) {
    checkRestrictionsForProgramExpr(loop.getCondition());
  }

  /**
   * Assignments have program expressions on their value part so restriction
   * on them must be checked.
   * 
   * @param assgnt the assignment instruction to check the restrictions for its
   *   value part
   */
  @Check
  public void checkProgramRestrictionsForAssignment(final Assignment assgnt) {
    checkRestrictionsForProgramExpr(assgnt.getValue());
  }
  
  /**
   * Expressions that are program expressions are restricted when compared
   * with expressions that may occur in specifications. The restrictions that
   * are checked here are:
   * <ul>
   * <li> quantifiers are not allowed in expressions</li>
   * <li> predicate calls are not allowed in expressions</li>
   * <li> counter variables are not allowed in expressions</li>
   * </ul> 
   * @param expr the expression to check the conditions for
   */
  private void checkRestrictionsForProgramExpr(final Expression expr) {
    // TODO: This is too simplistic - but works for now.
    final Iterator<EObject> iterator = 
        Iterators.singletonIterator((EObject)expr);
    while (iterator.hasNext()) {
      final EObject subexpr = iterator.next();
      if (subexpr instanceof Forall || subexpr instanceof Exists) {
        error("Quantifiers are not allowed in expressions.", 
              subexpr.eContainer(), subexpr.eContainingFeature(), -1); 
      } else if (subexpr instanceof Call) {
        final Call call = (Call)subexpr;
        if (call.getFunction() instanceof Predicate) {
          error("Predicate calls are not allowed in expressions.", 
              call, Literals.CALL__FUNCTION, -1);           
        }
      } else if (subexpr instanceof Var) {
        final Var var = (Var)subexpr;
          if (var.getSymbol() instanceof CounterVariable) {
          error("Counter variables are not allowed in expressions.",
                var, Literals.VAR__SYMBOL, -1); 
        }
      }
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////
  //// Type checking
  ////////////////////////////////////////////////////////////////////////////
  @Check
  public void checkAssignmentTypes(final Assignment assignment) {
    final Type type = getLValueType(assignment.getTarget());
    if (type != null) {
      checkExpressionType(assignment.getValue(), type);
    }
  }

  @Check
  public void checkAssertionType(final Assertion assertion) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(assertion.getFormula(), bool);
  }

  @Check
  public void checkPreconditionType(final Precondition pre) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(pre.getFormula(), bool);
  }

  @Check
  public void checkPostonditionType(final Postcondition post) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(post.getFormula(), bool);
  }
  
  @Check
  public void checkInvariantType(final LoopInvariant inv) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(inv.getFormula(), bool);
  }

  @Check
  public void checkCounterFormulaType(final LoopCounter counter) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(counter.getFormula(), bool);
  }
  
  @Check
  public void checkCounterNewValueType(final LoopCounter counter) {
    checkExpressionType(counter.getNewValue(), 
                        counter.getVariable().getType());
  }
  
  @Check
  public void checkCondType(final Cond cond) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(cond.getCondition(), bool);
  }
  
  @Check
  public void checkLoopCondType(final Loop loop) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(loop.getCondition(), bool);
  }
  
  @Check
  public void checkPredicateType(final Predicate predicate) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(predicate.getFormula(), bool);
  }
  
  @Check
  public void checkAxiomType(final Axiom axiom) {
    final SimpleType bool = typeUtils.getBoolean();
    checkExpressionType(axiom.getFormula(), bool);
  }
  
  ////////////////////////////////////////////////////////////////////////////
  //// Helper type checking methods
  ////////////////////////////////////////////////////////////////////////////

  
  private void checkExpressionType(final Expression expr, final Type type) {
    new TypeChecker(type).doSwitch(expr);
  }
  
  private Type getLValueType(final LValue lvalue) {
    Type type = lvalue.getVariable().getType();
    
    if (type == null) { return null; }
    int depth = 0;
    for (final Expression index : lvalue.getPath()) {
      if (! (type instanceof ArrayType)) {
        error(
            String.format("%s : %s is not an array", 
               lvalue.getVariable().getName() + repeat("[]", depth), 
               type.asString()),
            lvalue, HahaPackage.Literals.LVALUE__VARIABLE, 0);
        return null;
      }
      final ArrayType arrayType = (ArrayType)type;
      final SimpleType z = typeUtils.getZ();
      checkExpressionType(index, z);   
      type = arrayType.getElementType();
      depth++;
    }
    return type;
  }
  
  private final String repeat(final String text, final int count) {
    final StringBuilder builder = new StringBuilder(text.length() * count);
    for (int i = 0; i < count; i++) {
      builder.append(text);
    }
    return builder.toString();
  }

  ////////////////////////////////////////////////////////////////////////////
  //// TypeChecker class
  ////////////////////////////////////////////////////////////////////////////
  private final class TypeChecker extends HahaSwitch<Object> {
    private final Object OK = this;

    private final Type type;
    private Var ugly;

    private int featureIndex = -1;
    
    private boolean isIntOrZ(final Type type) {
      if (type instanceof SimpleType) {
        final PrimitiveType ptype = ((SimpleType)type).getType();
        return ptype == PrimitiveType.Z || ptype == PrimitiveType.INT;
      } else {
        return false;
      }
    }
    
    private boolean isBool(final Type type) {
      if (type instanceof SimpleType) {
        final PrimitiveType ptype = ((SimpleType)type).getType();
        return ptype == PrimitiveType.BOOLEAN;
      } else {
        return false;
      }
    }
   
    private boolean isZ(final Type type) {
      if (type instanceof SimpleType) {
        final PrimitiveType ptype = ((SimpleType)type).getType();
        return ptype == PrimitiveType.Z;
      } else {
        return false;
      }
    }
    
    public TypeChecker(final Type type) {
      this.type = type;
    }

    public void setFeatureIndex(final int featureIndex) {
      this.featureIndex = featureIndex;
    }
    
    @Override
    public Object caseVar(final Var var) {
      ugly = var;
      doSwitch(var.getSymbol());
      return OK;
    }

    @Override
    public Object caseCall(final Call call) {
      final ActualArgs args = call.getArgs();
      final Callable var = call.getFunction();
      if (var instanceof Function) {
        final Function function = (Function) var;
        checkFuncCall(call, function, args);
      } else if (var instanceof Predicate) {
        final Predicate predicate = (Predicate)var;
        checkPredCall(call, predicate, args);
      } else {
        error(
            String.format("%s is neither a function nor a predicate", 
              var.getName()),
            call,
            HahaPackage.Literals.CALL__FUNCTION,  featureIndex);

      }

      return OK;
    }

    private void checkFuncCall(final Call call, final Function function,
        final ActualArgs args) {
      if (! typeUtils.typesAreEqual(type, function.getResultType())) {
        error(String.format("Function %s returns %s, not %s",
            function.getName(), 
            function.getResultType().asString(), 
            type.asString()), call,
            HahaPackage.Literals.CALL__FUNCTION,  featureIndex);
      }
      checkActualArgs(call, function.getArguments(), args);
    }

    private void checkPredCall(final Call call, final Predicate pred,
        final ActualArgs args) {
      checkActualArgs(call, pred.getArguments(), args);
    }

    private void checkActualArgs(final Call call, 
        final ArgumentList expected,
        final ActualArgs actual) {
      final int expectedArity = expected.getArguments().size();
      final int actualArity = actual.getArguments().size();
      if (expectedArity != actualArity) {
        error(String.format("Function %s expects %d arguments, not %d",
            call.getFunction().getName(), 
            expectedArity, actualArity), call,
            HahaPackage.Literals.CALL__FUNCTION, featureIndex);
      }

      final Iterator<Argument> i = expected.getArguments().iterator();
      final Iterator<Expression> j = actual.getArguments().iterator();
      int argIndex = 0;
      while (i.hasNext() && j.hasNext()) {
        final Argument expectedArg = i.next();
        final Expression actualArg = j.next();
        final TypeChecker argChecker = 
            new TypeChecker(expectedArg.getType());
        argChecker.setFeatureIndex(argIndex);
        argChecker.doSwitch(actualArg);
        argIndex = argIndex + 1;
        // TODO: Extract result, print arg name and index in error.
      }
    }
    
    @Override
    public Object caseVariable(final Variable var) {
      if (! typeUtils.typesAreEqual(type, var.getType())) {
        error(
            String.format("Variable %s has type %s, %s expected", 
              var.getName(),
              var.getType().asString(), 
              type.asString()
            ), ugly,
            HahaPackage.Literals.VAR__SYMBOL, featureIndex);
      }
      return OK;
    }
    
    @Override
    public Object caseUnaryArithOp(final UnaryArithOp expr) {
      if (! isIntOrZ(type)) {
        error(String.format("%s expected, number found", type.asString()),
            expr.eContainer(), expr.eContainingFeature(), featureIndex);    
      } else {
        doSwitch(expr.getArg());
      }
      return OK;
    }
    
    @Override
    public Object caseUnaryBoolOp(final UnaryBoolOp expr) {
      if (! isBool(type)) {
        error(String.format("%s expected, number found", type.asString()),
            expr.eContainer(), expr.eContainingFeature(), featureIndex);
      } else {
        doSwitch(expr.getArg());
      }
      return OK;
    }

    @Override
    public Object caseBinaryArithOp(final BinaryArithOp expr) {
      if (! isIntOrZ(type)) {
        error(String.format("%s expected, number found", type.asString()),
            expr.eContainer(), expr.eContainingFeature(), featureIndex);
      } else {
        doSwitch(expr.getLeft());
        doSwitch(expr.getRight());
      }
      return OK;
    }
    
    @Override
    public Object caseBinaryBoolOp(final BinaryBoolOp expr) {
      if (! isBool(type)) {
//        error(String.format("%s expected, boolean found", type.asString()),
//            expr.eContainer(), expr.eContainingFeature(), 0);
      error(String.format("%s expected, boolean found", type.asString()),
        expr, null, featureIndex);
      } else {
        doSwitch(expr.getLeft());
        doSwitch(expr.getRight());
      }
      return OK;
    }
    
    @Override
    public Object caseComparisonOp(final ComparisonOp expr) {
      if (! isBool(type)) {
        error(String.format("%s expected, boolean found", type.asString()),
            expr.eContainer(), expr.eContainingFeature(), featureIndex);
      } else {
        // TODO: Fix, replace types with constraints.
        final SimpleType z = typeUtils.getZ();
        final TypeChecker exprChecker = new TypeChecker(z);
        exprChecker.doSwitch(expr.getLeft());
        exprChecker.doSwitch(expr.getRight());
      }
      return OK;
    }
    
    @Override
    public Object caseArrayAccess(final ArrayAccess a) {
      Type expectedType = type;
      final SimpleType z = typeUtils.getZ();
      final TypeChecker indexChecker = new TypeChecker(z);
      int depth = 0;
      for (final Expression index : a.getPath()) {
        indexChecker.setFeatureIndex(depth);
        indexChecker.doSwitch(index);
        expectedType = typeUtils.getArray(type);
        depth++;
      }
      final TypeChecker checker = new TypeChecker(expectedType);
      checker.doSwitch(a.getArray());
      return OK;
    }
    
    @Override
    public Object caseIntLiteral(final IntLiteral literal) {
      if (! isZ(type)) {
        error(String.format("%s expected, integer found", type.asString()), 
            literal.eContainer(), literal.eContainingFeature(), featureIndex);
      }
      return OK;
    }

    @Override
    public Object caseBoolLiteral(final BoolLiteral literal) {
      if (! isBool(type)) {
        error(String.format("%s expected, boolean found", type.asString()), 
            literal.eContainer(), literal.eContainingFeature(), featureIndex);
      }
      return OK;
    }
    
  }
  
  ////////////////////////////////////////////////////////////////////////////
  // Switch used to check for unassigned function result
  ////////////////////////////////////////////////////////////////////////////
  private final class ResultChecker extends HahaSwitch<Boolean> {

    @Override
    public Boolean caseHoareTriple(final HoareTriple triple) {
      return doSwitch(triple.getStatement());
    }
    
    @Override
    public Boolean caseBlock(final Block block) {
      for (final Statement stmt : block.getStatements()) {
        if (doSwitch(stmt)) {
          return true;
        }
      }
      return false;
    }

    @Override
    public Boolean caseSkip(final Skip skip) {
      return false;
    }

    @Override
    public Boolean caseAssignment(final Assignment assignment) {
      return assignment.getTarget().getVariable() instanceof ResultVar;
    }

    @Override
    public Boolean caseCond(final Cond cond) {
        if (cond.getIfFalse() == null) {
          return false;
        }
        return doSwitch(cond.getIfTrue()) && doSwitch(cond.getIfFalse());
    }

    @Override
    public Boolean caseLoop(final Loop object) {
        return false;
    }
       
  }  
}
