package net.afyre.haha.demo.vcgen;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.nodemodel.ICompositeNode;
import org.eclipse.xtext.nodemodel.util.NodeModelUtils;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;

import net.afyre.haha.demo.haha.Assertion;
import net.afyre.haha.demo.haha.AssertionList;
import net.afyre.haha.demo.haha.Block;
import net.afyre.haha.demo.haha.Cond;
import net.afyre.haha.demo.haha.HoareTriple;
import net.afyre.haha.demo.haha.Loop;
import net.afyre.haha.demo.haha.LoopCounter;
import net.afyre.haha.demo.haha.LoopInvariant;
import net.afyre.haha.demo.haha.SimpleStatement;
import net.afyre.haha.demo.haha.Statement;
import net.afyre.haha.demo.haha.util.HahaSwitch;
import net.afyre.haha.demo.solver.context.Declaration;
import net.afyre.haha.demo.solver.context.TermDeclaration;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.formulae.impl.Conjunction;
import net.afyre.haha.demo.solver.expr.formulae.impl.Exists;
import net.afyre.haha.demo.solver.expr.formulae.impl.Ge;
import net.afyre.haha.demo.solver.expr.formulae.impl.Gt;
import net.afyre.haha.demo.solver.expr.formulae.impl.Lt;
import net.afyre.haha.demo.solver.expr.formulae.impl.Negation;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.terms.impl.IntConst;
import net.afyre.haha.demo.solver.expr.terms.impl.Variable;
import net.afyre.haha.demo.solver.expr.util.Substitution;
import net.afyre.haha.demo.util.MappedIterable;

/**
 * VCGen implementation for single statements.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class DefaultHoareTripleVCGen implements HoareTripleVCGen {
  /** Object used to convert expressions to documented formulae. */
  private final ExprToDocumentedFormula exprToFormula = 
      new ExprToDocumentedFormula();

  /** Object used to convert assertions to documented formulae. */
  private final AssertionToDocumentedFormula assertToFormula = 
      new AssertionToDocumentedFormula();
  
  /** Utility methods. */
  private final VCGenUtil util = VCGenUtil.getInstance();
  
  /** {@inheritDoc} */
  @Override
  public void generateVerificationConditions(
      final VCContext context,
      final List<DocumentedFormula> preconditions,
      final HoareTriple stmt,
      final List<DocumentedFormula> postconditions,
      final VCConsumer consumer) {
    // Check if triple preconditions are valid.
    final VCContext preContext = new VCContext(
        "pre_valid", ImmutableList.of((EObject)stmt), 
        "Precondition correctness", context);
    for (final Assertion pre : stmt.getPrecondition().getAssertions()) {
      final DocumentedFormula preFormula = assertToFormula.apply(pre);
      final VerificationCondition vc = 
          new VerificationCondition(preconditions, preFormula, preContext);
      consumer.processVerificationCondition(vc);
    }
    
    // Compute actual pre- and post-conditions for the statement.
    final List<DocumentedFormula> preFormulae;
    if (stmt.getPrecondition().getAssertions().isEmpty()) {
      preFormulae = preconditions;
    } else {
      preFormulae = Lists.newArrayList(
        MappedIterable.getInstance(stmt.getPrecondition().getAssertions(), 
                                   assertToFormula)                                 
      );
    }
    final List<DocumentedFormula> postFormulae;
    if (stmt.getPostcondition().getAssertions().isEmpty()) {
      postFormulae = postconditions;
    } else {
      postFormulae = Lists.newArrayList(
        MappedIterable.getInstance(stmt.getPostcondition().getAssertions(), 
                                   assertToFormula)                                 
      );
    }
    
    final SimpleStatementVCGen simpleGen = new DefaultSimpleStatementVCGen();
    final Switch worker = 
        new Switch(simpleGen, 
                   context, preFormulae, postFormulae, 
                   consumer);
    if (!worker.doSwitch(stmt.getStatement())) {
      throw new RuntimeException("Unsupported statement: " + 
                    stmt.getStatement().eClass().getName());
    }
    
    // Check postconditions.
    if (!stmt.getPostcondition().getAssertions().isEmpty()) {
      final VCContext postContext = new VCContext(
           "post_valid", ImmutableList.of((EObject)stmt), 
           "Postcondition correctness.", context);
      for (final DocumentedFormula post : postconditions) {
        final VerificationCondition vc = 
            new VerificationCondition(postFormulae, post, postContext);
        consumer.processVerificationCondition(vc);
      }
    }
  }

  private final class Switch extends HahaSwitch<Boolean> {
    /** Processor for simple statements. */
    private final SimpleStatementVCGen simpleGen;
    
    /** Parent context for generated conditions. */
    private final VCContext parentContext;
    
    /** List of preconditions. */
    private final List<DocumentedFormula> preconditions;
    
    /** List of postconditions. */
    private final List<DocumentedFormula> postconditions;
  
    /** Object used to process generated conditions. */
    private final VCConsumer consumer;  
    
    /** Translates AST types to SMT types. */
    private final ExpressionUtil exprUtil = ExpressionUtil.getInstance();
    
    /**
     * Constructor.
     * @param simpleGen Processor for simple statements.
     * @param parentContext Parent context for generated conditions.
     * @param preconditions List of preconditions.
     * @param postconditions List of postconditions.
     * @param consumer Object used to process generated conditions.
     */
    public Switch(final SimpleStatementVCGen simpleGen,
                  final VCContext parentContext,
                  final List<DocumentedFormula> preconditions,
                  final List<DocumentedFormula> postconditions,
                  final VCConsumer consumer) {
      this.simpleGen = simpleGen;
      this.parentContext = parentContext;
      this.preconditions = preconditions;
      this.postconditions = postconditions;
      this.consumer = consumer;
    }
    
    // Default case = unsupported statement.
    @Override
    public Boolean defaultCase(final EObject obj) {
      return false;
    }
    
    // Delegate basic statements to suitable VCGen.
    @Override
    public Boolean caseSimpleStatement(final SimpleStatement stmt) {
      simpleGen.generateVerificationConditions(
          parentContext, 
          preconditions, stmt, postconditions, 
          consumer);
      return true;
    }
    
    // Conditional statement
    @Override
    public Boolean caseCond(final Cond cond) {
      List<DocumentedFormula> extendedPreconditions;
      final DocumentedFormula condFormula;
      
      VCContext ifContext = 
          new VCContext(
              "cond_L" + getLine(cond), 
              ImmutableList.of((EObject)cond), 
              "If statement at line " + getLine(cond), 
              parentContext);
      ifContext = util.processFunctionCalls(ifContext, 
          preconditions, cond.getCondition(), consumer);
      condFormula = exprToFormula.apply(cond.getCondition());
      
      // Case 1 - condition holds.
      final VCContext trueContext = 
          new VCContext(
              "ifTrue", 
              ImmutableList.of((EObject)cond.getIfTrue()),
              "Case 1 - condition holds.",
              ifContext);
      extendedPreconditions = 
          new ArrayList<DocumentedFormula>(preconditions);
      extendedPreconditions.add(condFormula);
      
      generateVerificationConditions(
          trueContext,
          extendedPreconditions, 
          cond.getIfTrue(), 
          postconditions, 
          consumer);
      
      // Case 2 - condition does not hold.      
      final DocumentedFormula negatedCondition = 
          new DocumentedFormula(
              new Negation(condFormula.getFormula()), 
              "negated_condition", 
              ImmutableList.of((EObject)cond), 
              "Negated condition");
      
      extendedPreconditions = 
          new ArrayList<DocumentedFormula>(preconditions);
      extendedPreconditions.add(negatedCondition);

      final EObject falseSource;
      if (cond.getIfFalse() != null) {
        falseSource = cond.getIfFalse();
      } else {
        falseSource = cond;
      }
      final VCContext falseContext = 
          new VCContext("ifFalse", 
                        ImmutableList.of(falseSource),
                        "Case 2 - condition does not hold.",
                        ifContext);
      // If there is a 'false' branch, process recursively.
      if (cond.getIfFalse() != null) {
        generateVerificationConditions(
            falseContext,
            extendedPreconditions, 
            cond.getIfFalse(), 
            postconditions, 
            consumer);
      } else {
        // Otherwise, generate implications.
        for (final DocumentedFormula post : postconditions) {
          final VerificationCondition vc = new VerificationCondition(
              extendedPreconditions, post, falseContext);
          consumer.processVerificationCondition(vc);
        }
      }
      return true;
    }
    
    // Substitituion in documented formula.
    private DocumentedFormula substInDF(final Substitution subst, 
                                        final DocumentedFormula formula) {
      final Formula newFormula = subst.perform(formula.getFormula());
      return new DocumentedFormula(newFormula, formula.getName(), 
                                   formula.getSource(), formula.getComment());
    }
    
    private Formula nat(final String variable) {
      return new Ge(new Variable(variable), new IntConst(BigInteger.ZERO));
    }
    
    private DocumentedFormula counterValueExists(final LoopCounter counter) {
      final VariableBinding binding = 
          new VariableBinding(
              counter.getVariable().getName(), 
              exprUtil.translateType(counter.getVariable().getType()));
      final Formula formula = 
        new Exists(ImmutableList.of(binding), 
          new Conjunction(nat(
               counter.getVariable().getName()), 
               exprToFormula.apply(counter.getFormula()).getFormula())
        );
      return new DocumentedFormula(formula, 
                                   "loop_counter_exists", 
                                   ImmutableList.of((EObject)counter), 
                                   "Loop counter exists");
    }
    
    private DocumentedFormula counterAtZero(final LoopCounter counter) {
      final Substitution subst = new Substitution();
      final Term value = new IntConst(BigInteger.ZERO);
      subst.setTermBinding(counter.getVariable().getName(), value);
      return substInDF(subst, exprToFormula.apply(counter.getFormula()));  
    }
    
    private DocumentedFormula counterAtNewValue(final LoopCounter counter) {
      final Substitution subst = new Substitution();
      final Term value = exprUtil.translateTerm(counter.getNewValue());
      subst.setTermBinding(counter.getVariable().getName(), value);
      return substInDF(subst, exprToFormula.apply(counter.getFormula()));  
    }

    private DocumentedFormula counterIsPositive(final LoopCounter counter) {
      final Formula formula = new Conjunction(
        new Gt(new Variable(counter.getVariable().getName()), 
               new IntConst(BigInteger.ZERO)),
        exprToFormula.apply(counter.getFormula()).getFormula());
        return new DocumentedFormula(formula, "counter_is_positive", 
                                     ImmutableList.of((EObject)counter), 
                                     "Check if the counter is positive.");
    }
    
    private DocumentedFormula newValueValid(final LoopCounter counter) {
      final Term value = exprUtil.translateTerm(counter.getNewValue());
      final Formula isNat = new Ge(value, new IntConst(BigInteger.ZERO));
      final Formula isLt = new Lt(
          value, 
          new Variable(counter.getVariable().getName()));
      final Formula formula = new Conjunction(isNat, isLt);
      
      return new DocumentedFormula(
               formula, "new_counter_valid", 
               ImmutableList.of((EObject)counter.getNewValue()), 
               "New counter value validity.");
    }
    
    // While loop.
    @Override
    public Boolean caseLoop(final Loop loop) {
      final DocumentedFormula condFormula;
      final DocumentedFormula negatedCondFormula;
      final DocumentedFormula counterFormula;
      
      final List <Declaration> declarations =
          new ArrayList<Declaration>();
      
      if (loop.getCounter() != null) {
        counterFormula = exprToFormula.apply(loop.getCounter().getFormula());
        declarations.add(new TermDeclaration(
          loop.getCounter().getVariable().getName(), 
          exprUtil.translateType(loop.getCounter().getVariable().getType())));
      } else {
        counterFormula = null;
      }
      
      VCContext loopContext = 
          new VCContext(
              "loop_L" + getLine(loop), 
              ImmutableList.of((EObject)loop), 
              "Loop statement at line " + getLine(loop), 
              parentContext,
              declarations);
      loopContext = util.processFunctionCalls(loopContext, 
                                              preconditions,
                                              loop.getCondition(), 
                                              consumer);
      
      condFormula = exprToFormula.apply(loop.getCondition());
      negatedCondFormula = 
          new DocumentedFormula(
              new Negation(condFormula.getFormula()), 
              "negated_condition", 
              ImmutableList.of((EObject)loop.getCondition()), 
              "Negated condition");

      
      // Collect loop invariants.
      final List<DocumentedFormula> invariants = 
          new ArrayList<DocumentedFormula>();
      for (final LoopInvariant invariant : loop.getInvariants()) {
        invariants.add(exprToFormula.apply(invariant.getFormula()));
      }
      
      // Preconditions must imply invariants.
      final VCContext preContext = 
          new VCContext(
              "pre", 
              ImmutableList.of((EObject)loop), 
              "Invariants before the loop", 
              loopContext);      
      for (final DocumentedFormula inv : invariants) {
        final VerificationCondition vc = 
            new VerificationCondition(preconditions, inv, preContext);
        consumer.processVerificationCondition(vc);
      }
      
      if (loop.getCounter() != null) {
        // Preconditions must imply existence of a valid counter value.
        final VerificationCondition counterExists = 
            new VerificationCondition(preconditions, 
                                      counterValueExists(loop.getCounter()), 
                                      loopContext);
        consumer.processVerificationCondition(counterExists);
        
        // Counter must match the condition.
        final VerificationCondition counterIter = 
            new VerificationCondition(
              ImmutableList.of(counterIsPositive(loop.getCounter()),
                               counterFormula),
              condFormula, loopContext
            );
        consumer.processVerificationCondition(counterIter);
        
        final VerificationCondition counterExit = 
            new VerificationCondition(ImmutableList.of(
                                        counterAtZero(loop.getCounter())), 
                                      negatedCondFormula, loopContext);
        consumer.processVerificationCondition(counterExit);
      }
      
      // Invariants + negated condition must imply postconditions.
      final VCContext postContext = 
          new VCContext(
              "post", 
              ImmutableList.of((EObject)loop), 
              "Loop postcondition.", 
              loopContext);
      
      final List<DocumentedFormula> invariantsAndNotCond = 
          new ArrayList<DocumentedFormula>(invariants);  
      invariantsAndNotCond.add(negatedCondFormula);
      
      if (loop.getCounter() != null) {
        invariantsAndNotCond.add(counterAtZero(loop.getCounter()));
      }
      
      for (final DocumentedFormula post : postconditions) {
        final VerificationCondition vc = 
            new VerificationCondition(invariantsAndNotCond, post, postContext);
        consumer.processVerificationCondition(vc);
      }
      
      // Iteration must preserve invariants.
      final VCContext invContext = 
          new VCContext(
              "invariants", 
              ImmutableList.of((EObject)loop.getBody()), 
              "Invariants after a single iteration.", 
              loopContext);
      
      final List<DocumentedFormula> invariantsAndCond = 
          new ArrayList<DocumentedFormula>(invariants);
      final List<DocumentedFormula> postInvariants = 
          new ArrayList<DocumentedFormula>(invariants);
      invariantsAndCond.add(condFormula);  
      
      if (loop.getCounter() != null) {
        invariantsAndCond.add(counterFormula);
        invariantsAndCond.add(counterIsPositive(loop.getCounter()));
        postInvariants.add(newValueValid(loop.getCounter()));
        postInvariants.add(counterAtNewValue(loop.getCounter()));
      }
      
      generateVerificationConditions(
          invContext, 
          invariantsAndCond, loop.getBody(), postInvariants, 
          consumer);
      
      return true;
    }
    
    @Override
    public Boolean caseBlock(final Block block) {
      final VCContext blockContext = 
          new VCContext(
              "block_L" + getLine(block) + "_" + getEndLine(block), 
              ImmutableList.of((EObject)block), 
              "Block at lines " + getLine(block) + 
                                  " - " + getEndLine(block), 
              parentContext);
      List<DocumentedFormula> currentState = preconditions;
      final Iterator<Statement> stmtIter = block.getStatements().iterator();
      final Iterator<AssertionList> assertIter = 
          block.getMidconditions().iterator();
      while (assertIter.hasNext()) {
        assert stmtIter.hasNext();
        
        final Statement stmt = stmtIter.next();
        final AssertionList postAsserts = assertIter.next();
        final List<DocumentedFormula> postFormulae = 
            Lists.newArrayList(
              MappedIterable.getInstance(postAsserts.getAssertions(), 
                                         assertToFormula));
        final Switch sw = new Switch(simpleGen, blockContext, 
                                     currentState, postFormulae, 
                                     consumer);
        sw.doSwitch(stmt);
        currentState = postFormulae;
      }
    
      if (stmtIter.hasNext()) {
        // Process last statement.
        final Statement stmt = stmtIter.next();
        final Switch sw = new Switch(simpleGen, blockContext, 
                                     currentState, postconditions, 
                                     consumer);
        sw.doSwitch(stmt);
      } else {
        // The block was empty.
        for (final DocumentedFormula post : postconditions) {
          final VerificationCondition vc = new VerificationCondition(
              preconditions, post, blockContext);
          consumer.processVerificationCondition(vc);
        }
      }

      return true;
    }
      
    // Get line number from a statement.
    // Return 'X' if line number information is not available.
    private String getLine(final Statement stmt) {
      ICompositeNode node = NodeModelUtils.getNode(stmt);
      if (node == null) {
        return "X";
      } else {
        return Integer.toString(node.getStartLine());
      }
    }
    
    // Get ending line number from a statement.
    // Return 'X' if lnie number information is not available.
    private String getEndLine(final Statement stmt) {
      ICompositeNode node = NodeModelUtils.getNode(stmt);
      if (node == null) {
        return "X";
      } else {
        return Integer.toString(node.getEndLine());
      }
    }
    
  }
}
