package net.afyre.haha.demo.vcgen;

import java.util.List;

import net.afyre.haha.demo.solver.context.Declaration;

import org.eclipse.emf.ecore.EObject;

import com.google.common.collect.ImmutableList;

/**
 * The context of a verification condition is a tree structure
 * which describes its origin and visible symbol declarations.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class VCContext {
  /** Short name. */
  private final String name;
  
  /** Source elements associated with this context. */
  private final List<EObject> source;
  
  /** Descriptive comment. */
  private final String comment;
  
  /** Parent context (can be null). */
  private final VCContext parent;

  /** List of declarations visible in this context. */
  private final List<Declaration> declarations;
  
  /**
   * Create a context with empty list of declarations.
   * @param name Short name.
   * @param source Source elements.
   * @param comment Descriptive comment.
   * @param parent Parent context (can be null).
   */
  public VCContext(final String name, 
                   final List<? extends EObject> source, 
                   final String comment,
                   final VCContext parent) {
    this.name = name;
    this.source = ImmutableList.copyOf(source);
    this.comment = comment;
    this.parent = parent;
    this.declarations = ImmutableList.<Declaration>of();
  }

  /**
   * Create a context with given list of declarations.
   * @param name Short name.
   * @param source Source elements.
   * @param comment Descriptive comment.
   * @param parent Parent context (can be null).
   * @param declarations List of declarations.
   */
  public VCContext(final String name, 
                   final List<? extends EObject> source, 
                   final String comment,
                   final VCContext parent, 
                   final Iterable<? extends Declaration> declarations) {
    this.name = name;
    this.source = ImmutableList.copyOf(source);
    this.comment = comment;
    this.parent = parent;
    this.declarations = 
        ImmutableList.<Declaration>copyOf(declarations);
  }
  
  /**
   * Get the short name of this context.
   * @return Context name.
   */
  public String getName() {
    return name;
  }

  /**
   * Get list of source elements associated with this context.
   * Returned list is not modifiable.
   * @return List of AST elements.
   */
  public List<? extends EObject> getSource() {
    return source;
  }

  /**
   * Get the comment associated with this context.
   * @return Comment text.
   */
  public String getComment() {
    return comment;
  }

  /**
   * Get the parent context (null for root contexts).
   * @return Parent context or null.
   */
  public VCContext getParent() {
    return parent;
  }

  /**
   * Get list of declarations.
   * @return Declarations.
   */
  public Iterable<Declaration> getDeclarations() {
    return declarations;
  }
  
  /**
   * Check if this frame contains any declarations.
   * @return True iff the declaration list is not empty.
   */
  public boolean containsDeclaration() {
    return !declarations.isEmpty();
  }
  
}
