package net.afyre.haha.demo.vcgen;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;
import java.util.Stack;

/**
 * Abstract VC consumer which decouples handling of contexts
 * from handling of formulae.
 * 
 * It maintains a stack of contexts. Subclasses are notified
 * whenever a context is added to or removed from that stack 
 * ({@link #pushContext(VCContext)}, {@link #popContext(VCContext)}).
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public abstract class VCProcessor implements VCConsumer {
  /** Stack of contexts. */
  private final Stack<VCContext> contextStack = new Stack<VCContext>();
  
  /** Contexts in scope. */
  private final Set<VCContext> contexts = new HashSet<VCContext>();
  
  /** Aggregate of verification conditions. */
  private final Hashtable<String,VerificationCondition> vcs = new Hashtable<String,VerificationCondition>();
  
  /** {@inheritDoc} */
  @Override
  public final void processVerificationCondition(
      final VerificationCondition vc) {
    final VCContext context = vc.getContext();
    
    VCContext ancestor = context;
    while (ancestor != null && !contexts.contains(ancestor)) {
      ancestor = ancestor.getParent();
    }
 
    while (!contextStack.isEmpty() &&
           contextStack.peek() != ancestor) {
      final VCContext removedContext = contextStack.pop();
      contexts.remove(removedContext);
      popContext(removedContext);
    }
    
    // We might have to push parent contexts first.
    final Stack<VCContext> newContexts = new Stack<VCContext>();
    
    VCContext current = context;
    while (current != ancestor) {
      newContexts.push(current);
      current = current.getParent();
    }
    
    // NOTE: Order is reversed here.
    while (!newContexts.isEmpty()) {
      final VCContext newContext = newContexts.pop();
      contextStack.push(newContext);
      contexts.add(newContext);
      pushContext(newContext);
    }
    StringBuffer name = new StringBuffer("");
    boolean fst=true;
    for (Iterator<VCContext> i = contextStack.iterator(); i.hasNext();) {
		VCContext vcContext = (VCContext) i.next();
		if (fst) {
			name.append(vcContext.getName());
			fst=false;
		} else {
			name.append(".").append(vcContext.getName());
		}
	}
    vcs.put(name + "." + vc.getGoal().getName(),vc);
    
    processCondition(vc);
  }

  /** Pop all contexts. */
  public void end() {
    while (!contextStack.isEmpty()) {
      final VCContext removedContext = contextStack.pop();
      popContext(removedContext);      
    }
  }
  
  public Hashtable<String,VerificationCondition> getVCs() {
	  return vcs;
  }
  
  /**
   * Process a verification condition.
   * @param vc Condition.
   */
  protected abstract void processCondition(
      final VerificationCondition vc);
  
  /**
   * Method called when a new context is placed on the context stack.
   * @param context New context.
   */
  protected abstract void pushContext(VCContext context);
  
  /**
   * Method called when a context is removed from a context stack.
   * @param context Removed context.
   */
  protected abstract void popContext(VCContext context);
}
